#include <avr/io.h>
#include <avr/interrupt.h>
#define F_CPU 1000000UL
#include <util/delay.h>
#include "USI_TWI_Master.h"

#define MESSAGEBUF_SIZE   8
#define RTC_SLAVE_ADDRESS 0xDF	// bit[7:4]=RTC mode, bit[3:1] is fixed address 0b111, bit[0]=R/Wn
#define EPR_SLAVE_ADDRESS 0xAF	// bit[7:4]=EPR mode, bit[3:1] is fixed address 0b111, bit[0]=R/Wn
#define RTC_SECOND        0x00	// bit[7]=ST 1=OSC on, bit[6:4]=BCD sec  ten, bit[3:0]=BCD sec  one
#define RTC_MINUTE        0x01	// bit[7]=nc,          bit[6:4]=BCD min  ten, bit[3:0]=BCD min  one 
#define RTC_HOUR          0x02	// bit[7]=nc,6=12h,5=PM, bit[4]=BCD hour ten, bit[3:0]=BCD hour one 
#define RTC_CONTROL       0x07

#define L1_FRONT_TOP	  0x01
#define L2_FRONT_BOT	  0x02
#define L3_MID_TOP		  0x04
#define L4_MID_BOT		  0x08
#define L5_GUN_RIGHT	  0x10
#define L6_GUN_LEFT		  0x20
#define L56_GUNS		  0x30
#define L1_L6			  0x3f


void Display_Time(void);
void Display_Time_Hour(void);
void Display_Time_Minute(void);

// Global
unsigned char messageBuf[MESSAGEBUF_SIZE];
unsigned char t_hour,t_minute,t_second;
unsigned char counter;

//-------------------------------------------------------------------------------------------------
// Interrupt Service Routine for external INT0 on pin 9
//-------------------------------------------------------------------------------------------------
ISR(INT0_vect) {
	if (counter == 9) {
		counter=0;
	} else {
		counter++;
	}

 	PORTA &= ~L3_MID_TOP;
	_delay_ms(40);
	PORTA |=  L3_MID_TOP;

	if ((counter & 0x01) == 0x01) {
		PORTA &= ~L2_FRONT_BOT;
		_delay_ms(100);
		PORTA |=  L2_FRONT_BOT;
	}

	if (counter == 8) {
		Display_Time();
	}
}

//-------------------------------------------------------------------------------------------------
// Convert Decimal value to BCD coded value
//-------------------------------------------------------------------------------------------------
unsigned char DEC2BCD(unsigned char value) {
	return ((value/10)<<4) | (value%10);
}

//-------------------------------------------------------------------------------------------------
// Convert BCD coded value to Decimal value
//-------------------------------------------------------------------------------------------------
unsigned char BCD2DEC(unsigned char value) {
	return (value>>4)*10 + (value & 0x0F);
}

//-------------------------------------------------------------------------------------------------
// Write I2C: Select slave, Write on address <a2> the value <d1>
// Send: 1) Slave Address + Write bit, 2) register address, 3) data
//-------------------------------------------------------------------------------------------------
unsigned char I2C_Write(unsigned char a1, unsigned char a2, unsigned char d1) {
	messageBuf[0] = a1 & 0xFE; // Write: Clear bit[0]
	messageBuf[1] = a2;
	messageBuf[2] = d1;
	return USI_TWI_Start_Transceiver_With_Data(messageBuf,3);
}

//-------------------------------------------------------------------------------------------------
// Read I2C: Select slave, Read from address <a2> and return the data
// Send: 1) Slave Address + Write bit, 2) register address, 3) Repeated Start, 
//       4) Slave Address + Read bit, read data (set message size)
//       Data is found in messageBuf[1..n], n=lenght-1 
//-------------------------------------------------------------------------------------------------
unsigned char I2C_Read(unsigned char a1, unsigned char a2, unsigned char lenght) {
	unsigned char code;
	// Send Slave address + Write bit and send register address. Write: Clear bit[0]
	messageBuf[0] = a1 & 0xFE; 
	messageBuf[1] = a2;
	code = USI_TWI_Start_Transceiver_With_Data_Stop(messageBuf,2,0); // Do not generate a stop condition
	
	// Send Slave address + Read bit and read data. Read: Set bit[0]
	messageBuf[0] = a1 | 0x01;
	code = USI_TWI_Start_Transceiver_With_Data(messageBuf,1+lenght);
	return code;	
}

//-------------------------------------------------------------------------------------------------
// RTC: Set Hour
//-------------------------------------------------------------------------------------------------
unsigned char RTC_Set_Hour(unsigned char hour) {
	unsigned char data;
	data=0x40 | DEC2BCD(hour);		//0x40: 12H clock, time is PM (does not matter here)
	return I2C_Write(RTC_SLAVE_ADDRESS,RTC_HOUR,data);
}

//-------------------------------------------------------------------------------------------------
// RTC: Set Minute
//-------------------------------------------------------------------------------------------------
unsigned char RTC_Set_Minute(unsigned char minute) {
	unsigned char data;
	data=DEC2BCD(minute);
	return I2C_Write(RTC_SLAVE_ADDRESS,RTC_MINUTE,data);
}

//-------------------------------------------------------------------------------------------------
// RTC: Set Second
//-------------------------------------------------------------------------------------------------
unsigned char RTC_Set_Second(unsigned char second) {
	unsigned char data;
	data=0x80 | DEC2BCD(second);	// Keep OSC enabled (0x80)
	return I2C_Write(RTC_SLAVE_ADDRESS,RTC_SECOND,data);
}

//-------------------------------------------------------------------------------------------------
// RTC: Read Minute
//-------------------------------------------------------------------------------------------------
unsigned char RTC_Read_Minute(unsigned char *minute) {
	return I2C_Read(RTC_SLAVE_ADDRESS,RTC_MINUTE,1);	
}

//-------------------------------------------------------------------------------------------------
// RTC: Read Second
//-------------------------------------------------------------------------------------------------
unsigned char RTC_Read_Second(unsigned char *second) {
	unsigned char code;
	code = I2C_Read(RTC_SLAVE_ADDRESS,RTC_SECOND,1);	
	*second=BCD2DEC((messageBuf[1] & 0x7F)); // Remove the enable OSC bit[7]
	return code;
}

//-------------------------------------------------------------------------------------------------
// RTC: Read Time: Hour,Minute,Second
//-------------------------------------------------------------------------------------------------
unsigned char RTC_Read_Time(unsigned char *second, unsigned char *minute, unsigned char *hour) {
	unsigned char code;
	code = I2C_Read(RTC_SLAVE_ADDRESS,RTC_SECOND,3);	
	*second=BCD2DEC((messageBuf[1] & 0x7F)); // Remove non second bits
	*minute=BCD2DEC((messageBuf[2] & 0x7F)); // Remove non minute bits
	*hour  =BCD2DEC((messageBuf[3] & 0x1F)); // Remove non hour bits
	return code;
}

//-------------------------------------------------------------------------------------------------
// Display the Time: Hours, Quarters, Minutes
//-------------------------------------------------------------------------------------------------
void Display_Time(void) {
	unsigned char result, i, quarter, remain_min;
	
	//cli();
	result = RTC_Read_Time(&t_second, &t_minute, &t_hour);

	for (i=0; i<t_hour; i++) {
		PORTA &= ~L56_GUNS;	// on
		_delay_ms(200);
		PORTA |=  L56_GUNS;	// off
		_delay_ms(200);
	}

	quarter    = t_minute / 15;
	remain_min = t_minute % 15;
	
	_delay_ms(400);
	for (i=0; i<quarter; i++) {
		PORTA &= ~L1_FRONT_TOP;	// on
		_delay_ms(150);
		PORTA |=  L1_FRONT_TOP;	// off
		_delay_ms(100);
	} 

	_delay_ms(400);
	for (i=0; i<remain_min; i++) {
		PORTA &= ~L56_GUNS;	// on
		_delay_ms(200);
		PORTA |=  L56_GUNS;	// off
		_delay_ms(200);
	}
	//sei();
}

//-------------------------------------------------------------------------------------------------
// Display the Time: Hours
//-------------------------------------------------------------------------------------------------
void Display_Time_Hour(void) {
	unsigned char result, i;

	result = RTC_Read_Time(&t_second, &t_minute, &t_hour);

	for (i=0; i<t_hour; i++) {
		PORTA &= ~L56_GUNS;	// on
		_delay_ms(200);
		PORTA |=  L56_GUNS;	// off
		_delay_ms(100);
	}
}

//-------------------------------------------------------------------------------------------------
// Display the Time: Quarters, Minutes
//-------------------------------------------------------------------------------------------------
void Display_Time_Minute(void) {
	unsigned char i, quarter, remain_min;
	quarter    = t_minute / 15;
	remain_min = t_minute % 15;
	
	for (i=0; i<quarter; i++) {
		PORTA &= ~L1_FRONT_TOP;	// on
		_delay_ms(150);
		PORTA |=  L1_FRONT_TOP;	// off
		_delay_ms(100);
	} 

	for (i=0; i<remain_min; i++) {
		PORTA &= ~L56_GUNS;	// on
		_delay_ms(200);
		PORTA |=  L56_GUNS;	// off
		_delay_ms(100);
	}
}

//-------------------------------------------------------------------------------------------------
// MAIN ROUTINE
//-------------------------------------------------------------------------------------------------
int main (void) {	
	unsigned char result;
	
	counter = 0;

	//---------------------------------------------------------------------------------------------
	// Initialize
	// PA[7:6] is input (write 0), PA[5:0] is output (write 1)
	DDRA  = 0x3F;
	PORTA = 0xC0 | L1_L6;			// Value zero turns on a led
	PORTA &= ~L4_MID_BOT;

	// Setup interrupt handling
	GIMSK = 1<<INT0;				// Enable external interrupt
	MCUCR = 1<<ISC01 | 0<<ISC00;	// Falling edge on INT0 generates an interrupt
	sei();							// Global interrupt enable I<-1 in SREG

	// Setup I2C serial interface
	USI_TWI_Master_Initialise();

	// Setup Real Time Clock chip
	result = I2C_Write(RTC_SLAVE_ADDRESS,RTC_SECOND,0x80);		// Enable 32.768kHz oscillator
	result = I2C_Write(RTC_SLAVE_ADDRESS,RTC_CONTROL,0x40);		// Set output MPF to 1Hz pulse
	
	// Testcode: do not set this, it would overwrite the battery-backed-up time...
	//t_hour   = 0;
	//t_minute = 0;
	//t_second = 0;
	//result = RTC_Set_Hour(t_hour);
	//result = RTC_Set_Minute(t_minute);
	//result = RTC_Set_Second(t_second);
	//---------------------------------------------------------------------------------------------
	
	while(1) {	
		_delay_ms(200);

		// Check if S2=Minute button is pressed
		if ((PINA & 0x40) == 0) {
			if (t_minute < 59) {
				t_minute++;
			} else {
				t_minute=0;
			}
			RTC_Set_Minute(t_minute);
		}

		// Check if S1=Hour button is pressed
		if ((PINA & 0x80) == 0) {
			if (t_hour < 12) {
				t_hour++;
			} else {
				t_hour=0;
			}
			RTC_Set_Hour(t_hour);
		}		
	}
}
